from .getter import (
    get_base_info,
    get_fund_codes,
    get_industry_distribution,
    get_invest_position,
    get_pdf_reports,
    get_period_change,
    get_public_dates,
    get_quote_history,
    get_quote_history_multi,
    get_realtime_increase_rate,
    get_types_percentage,
    get_fund_manager,
)

__all__ = [
    "get_base_info",
    "get_fund_codes",
    "get_realtime_increase_rate",
    "get_invest_position",
    "get_types_percentage",
    "get_period_change",
    "get_public_dates",
    "get_industry_distribution",
    "get_quote_history",
    "get_quote_history_multi",
    "get_pdf_reports",
    "get_fund_manager",
]



import json
from pathlib import Path
from typing import Dict

import pandas as pd
import requests

from ..config import SEARCH_RESULT_CACHE_PATH, MAX_CONNECTIONS
from .tickflow_prompt import session

adapter = requests.adapters.HTTPAdapter(
    pool_connections=MAX_CONNECTIONS, pool_maxsize=MAX_CONNECTIONS, max_retries=5
)
session.mount("http://", adapter)
session.mount("https://", adapter)

# 关键词搜索缓存
SEARCH_RESULT_DICT: Dict[str, dict] = dict()
# 行情ID搜索缓存
BASE_INFO_CACHE: Dict[str, pd.Series] = dict()
path = Path(SEARCH_RESULT_CACHE_PATH)
if path.exists():
    load_success = False
    with path.open("r", encoding="utf-8") as f:
        try:
            SEARCH_RESULT_DICT = json.load(f)
            load_success = True
        except:
            pass
    if not load_success:
        path.open("w").close()



from .getter import (
    get_all_company_performance,
    get_all_report_dates,
    get_base_info,
    get_belong_board,
    get_daily_billboard,
    get_deal_detail,
    get_history_bill,
    get_latest_holder_number,
    get_latest_ipo_info,
    get_latest_quote,
    get_members,
    get_quote_history,
    get_quote_snapshot,
    get_realtime_quotes,
    get_today_bill,
    get_top10_stock_holder_info,
)

__all__ = [
    "get_history_bill",
    "get_today_bill",
    "get_latest_quote",
    "get_quote_history",
    "get_realtime_quotes",
    "get_top10_stock_holder_info",
    "get_base_info",
    "get_all_report_dates",
    "get_all_company_performance",
    "get_latest_holder_number",
    "get_daily_billboard",
    "get_members",
    "get_latest_ipo_info",
    "get_quote_snapshot",
    "get_deal_detail",
    "get_belong_board",
]



from pathlib import Path

HERE = Path(__file__).parent
FUTURES_BASE_INFO_SAVE_PATH = HERE / "futures_info.csv"
# K 线表头映射
EASTMONEY_KLINE_FIELDS = {
    "f51": "日期",
    "f52": "开盘",
    "f53": "收盘",
    "f54": "最高",
    "f55": "最低",
    "f56": "成交量",
    "f57": "成交额",
    "f58": "振幅",
    "f59": "涨跌幅",
    "f60": "涨跌额",
    "f61": "换手率",
}

EASTMONEY_REQUEST_HEADERS = {
    "User-Agent": "Mozilla/5.0 (Windows NT 6.3; WOW64; Trident/7.0; Touch; rv:11.0) like Gecko",
    "Accept": "*/*",
    "Accept-Language": "zh-CN,zh;q=0.8,zh-TW;q=0.7,zh-HK;q=0.5,en-US;q=0.3,en;q=0.2",
    "Referer": "http://quote.eastmoney.com/center/gridlist.html",
}



EastmoneyRank = {
    "f12": "代码",
    "f14": "名称",
    "f3": "涨跌幅",
    "f15": "最新价",
    "f4": "涨跌额",
    "f5": "成交量",
    "f6": "成交额",
    "f18": "昨日收盘",
}

EastmoneyHeaders = {
    "Host": "19.push2.eastmoney.com",
    "User-Agent": "Mozilla/5.0 (Windows NT 6.3; WOW64; Trident/7.0; Touch; rv:11.0) like Gecko",
    "Accept": "*/*",
    "Accept-Language": "zh-CN,zh;q=0.8,zh-TW;q=0.7,zh-HK;q=0.5,en-US;q=0.3,en;q=0.2",
    "Referer": "http://quote.eastmoney.com/center/gridlist.html",
}
EastmoneyFundHeaders = {
    "User-Agent": "EMProjJijin/6.2.8 (iPhone; iOS 13.6; Scale/2.00)",
    "GTOKEN": "98B423068C1F4DEF9842F82ADF08C5db",
    "clientInfo": "ttjj-iPhone10,1-iOS-iOS13.6",
    "Content-Type": "application/x-www-form-urlencoded",
    "Host": "fundmobapi.eastmoney.com",
    "Referer": "https://mpservice.com/516939c37bdb4ba2b1138c50cf69a2e1/release/pages/FundHistoryNetWorth",
}



from pathlib import Path

HERE = Path(__file__).parent
# 数据缓存文件存储目录
DATA_DIR = HERE / "../data"
# 创建数据缓存文件目录
DATA_DIR.mkdir(parents=True, exist_ok=True)
# 搜索词缓存位置
SEARCH_RESULT_CACHE_PATH = str(DATA_DIR / "search-cache.json")

MAX_CONNECTIONS = 50

# 网络连接错误时是否展示 TickFlow 推广提示（可设为 False 关闭）
SHOW_TICKFLOW_PROMPT = True



import sys
import threading

import requests
from rich.console import Console

from ..config import SHOW_TICKFLOW_PROMPT

TICKFLOW_URL = "https://tickflow.org?utm_source=efinance"

_tickflow_after_traceback = threading.local()
_ipython_tickflow_shown = False


def is_connection_error_show_tickflow(exc: requests.exceptions.ConnectionError) -> bool:
    msg = str(exc).lower()
    return "max retries exceeded" in msg or "remote end closed connection" in msg


def _print_tickflow_prompt():
    console = Console()
    console.print()
    console.print(
        "[bold red]网络连接异常[/bold red]，可尝试使用 [link={}]TickFlow[/link] 获取更稳定的数据。".format(
            TICKFLOW_URL
        )
    )
    console.print("  → {}".format(TICKFLOW_URL))
    console.print()


_original_excepthook = sys.excepthook


def _excepthook(exc_type, exc_value, tb):
    _original_excepthook(exc_type, exc_value, tb)
    if (
        exc_type is requests.exceptions.ConnectionError
        and SHOW_TICKFLOW_PROMPT
        and exc_value is not None
        and is_connection_error_show_tickflow(exc_value)
    ):
        _print_tickflow_prompt()


def _install_excepthook():
    if sys.excepthook is not _excepthook:
        global _original_excepthook
        _original_excepthook = sys.excepthook
        sys.excepthook = _excepthook


def _install_ipython_exc():
    try:
        from IPython import get_ipython

        ip = get_ipython()
    except Exception:
        return
    if ip is None:
        return

    def _ipython_maybe_show_tickflow():
        global _ipython_tickflow_shown
        if _ipython_tickflow_shown:
            return
        if not SHOW_TICKFLOW_PROMPT:
            return
        exc = getattr(sys, "last_value", None)
        if (
            exc is not None
            and isinstance(exc, requests.exceptions.ConnectionError)
            and is_connection_error_show_tickflow(exc)
        ):
            _ipython_tickflow_shown = True
            _print_tickflow_prompt()

    def pre_execute():
        global _ipython_tickflow_shown
        _ipython_tickflow_shown = False

    def post_run_cell(result):
        if result.error_in_exec is not None:
            _tickflow_after_traceback.show = False
            _ipython_maybe_show_tickflow()

    def post_execute():
        if getattr(_tickflow_after_traceback, "show", False):
            _tickflow_after_traceback.show = False
            _ipython_maybe_show_tickflow()

    ip.events.register("pre_execute", pre_execute)
    ip.events.register("post_run_cell", post_run_cell)
    ip.events.register("post_execute", post_execute)


class CustomedSession(requests.Session):
    def request(self, *args, **kwargs):
        kwargs.setdefault("timeout", 180)
        try:
            return super(CustomedSession, self).request(*args, **kwargs)
        except requests.exceptions.ConnectionError as e:
            if SHOW_TICKFLOW_PROMPT and is_connection_error_show_tickflow(e):
                _tickflow_after_traceback.show = True
            raise


session = CustomedSession()

_install_excepthook()
_install_ipython_exc()



from pathlib import Path

HERE = Path(__file__).parent

# 股票基本信息表头
EASTMONEY_STOCK_BASE_INFO_FIELDS = {
    "f57": "股票代码",
    "f58": "股票名称",
    "f162": "市盈率(动)",
    "f167": "市净率",
    "f127": "所处行业",
    "f116": "总市值",
    "f117": "流通市值",
    "f198": "板块编号",
    "f173": "ROE",
    "f187": "净利率",
    "f105": "净利润",
    "f186": "毛利率",
}

EASTMONEY_STOCK_DAILY_BILL_BOARD_FIELDS = {
    "SECURITY_CODE": "股票代码",
    "SECURITY_NAME_ABBR": "股票名称",
    "TRADE_DATE": "上榜日期",
    "EXPLAIN": "解读",
    "CLOSE_PRICE": "收盘价",
    "CHANGE_RATE": "涨跌幅",
    "TURNOVERRATE": "换手率",
    "BILLBOARD_NET_AMT": "龙虎榜净买额",
    "BILLBOARD_BUY_AMT": "龙虎榜买入额",
    "BILLBOARD_SELL_AMT": "龙虎榜卖出额",
    "BILLBOARD_DEAL_AMT": "龙虎榜成交额",
    "ACCUM_AMOUNT": "市场总成交额",
    "DEAL_NET_RATIO": "净买额占总成交比",
    "DEAL_AMOUNT_RATIO": "成交额占总成交比",
    "FREE_MARKET_CAP": "流通市值",
    "EXPLANATION": "上榜原因",
}



import json
import re
import time
from collections import OrderedDict, namedtuple
from functools import wraps
from typing import Any, Callable, Dict, List, TypeVar, Union

import pandas as pd
import rich
from retry.api import retry

from ..common.config import FS_DICT, MARKET_NUMBER_DICT, MagicConfig, MarketType
from ..config import SEARCH_RESULT_CACHE_PATH
from ..shared import SEARCH_RESULT_DICT, session

# 函数变量
F = TypeVar("F")


def to_numeric(func: F) -> F:
    """
    将 DataFrame 或者 Series 尽可能地转为数字的装饰器

    Parameters
    ----------
    func : Callable
        返回结果为 DataFrame 或者 Series 的函数

    Returns
    -------
    Union[DataFrame, Series]

    """

    ignore = [
        "股票代码",
        "基金代码",
        "代码",
        "市场类型",
        "市场编号",
        "债券代码",
        "行情ID",
        "正股代码",
    ]

    @wraps(func)
    def run(*args, **kwargs):
        values = func(*args, **kwargs)
        if isinstance(values, pd.DataFrame):
            for column in values.columns:
                if column not in ignore:

                    values[column] = values[column].apply(convert)
        elif isinstance(values, pd.Series):
            for index in values.index:
                if index not in ignore:

                    values[index] = convert(values[index])
        return values

    def convert(o: Union[str, int, float]) -> Union[str, float, int]:
        if not re.findall("\d", str(o)):
            return o
        try:
            if str(o).isalnum():
                o = int(o)
            else:
                o = float(o)
        except:
            pass
        return o

    return run


# 存储证券代码的实体
Quote = namedtuple(
    "Quote",
    [
        "code",
        "name",
        "pinyin",
        "id",
        "jys",
        "classify",
        "market_type",
        "security_typeName",
        "security_type",
        "mkt_num",
        "type_us",
        "quote_id",
        "unified_code",
        "inner_code",
    ],
)


@retry(tries=3, delay=1)
def get_quote_id(
    stock_code: str,
    market_type: Union[MarketType, None] = None,
    use_local=True,
    suppress_error=False,
    **kwargs,
) -> str:
    """
    生成东方财富股票专用的行情ID

    Parameters
    ----------
    stock_code : str
        证券代码或者证券名称
    market_type : MarketType, optional
        市场类型，目前可筛选A股，港股，美股和英股。默认不筛选
    use_local : bool, optional
        是否使用本地缓存
    suppress_error : bool, optional
        遇到错误的股票代码，是否不报错，返回空字符串

    Returns
    -------
    str
        东方财富股票专用的 secid
    """
    if len(str(stock_code).strip()) == 0:
        if suppress_error:
            return ""
        raise Exception("证券代码应为长度不应为 0")
    quote = search_quote(
        stock_code, market_type=market_type, use_local=use_local, **kwargs
    )
    if isinstance(quote, Quote):
        return quote.quote_id
    if quote is None:
        if not suppress_error:
            rich.print(f'证券代码 "{stock_code}" 可能有误')
        return ""


def search_quote(
    keyword: str,
    market_type: Union[MarketType, None] = None,
    count: int = 1,
    use_local: bool = True,
    **kwargs,
) -> Union[Quote, None, List[Quote]]:
    """
    根据关键词搜索以获取证券信息

    Parameters
    ----------
    keyword : str
        搜索词(股票代码、债券代码甚至证券名称都可以)
    market_type : MarketType, optional
        市场类型，目前可筛选A股，港股，美股和英股。默认不筛选
    count : int, optional
        最多搜索结果数, 默认为 `1`
    use_local : bool, optional
        是否使用本地缓存

    Returns
    -------
    Union[Quote, None, List[Quote]]

    """
    # NOTE 本地仅存储第一个搜索结果
    if use_local and count == 1:
        quote = search_quote_locally(keyword, market_type=market_type)
        if quote:
            return quote
    url = "https://searchapi.eastmoney.com/api/suggest/get"
    params = (
        ("input", f"{keyword}"),
        ("type", "14"),
        ("token", "D43BF722C8E33BDC906FB84D85E326E8"),
        ("count", f"{max(count, 5)}"),
    )
    try:
        json_response = session.get(url, params=params).json()
        items = json_response["QuotationCodeTable"]["Data"]
    except json.JSONDecodeError:
        RuntimeWarning(
            "unable to parse search quote result, consider if you are blocked"
        )
        return None

    if items is not None and items:
        quotes = [
            Quote(*item.values())
            for item in items
            # 支持精确查找股票代码
            if (
                not kwargs.get(MagicConfig.QUOTE_SYMBOL_MODE, False)
                or (keyword == item["Code"])
            )
            # 支持筛选股票市场
            and (market_type is None or (market_type.value == item["Classify"]))
        ]
        # NOTE 暂时仅存储第一个搜索结果
        save_search_result(keyword, quotes[:1])
        if count == 1:
            return quotes[0] if len(quotes) >= 1 else None

        return quotes[:count]

    return None


def search_quote_locally(
    keyword: str, market_type: Union[MarketType, None] = None
) -> Union[Quote, None]:
    """
    在本地里面使用搜索记录进行关键词搜索

    Parameters
    ----------
    keyword : str
        搜索词
    market_type : MarketType, optional
        市场类型，目前可筛选A股，港股，美股和英股。默认不筛选

    Returns
    -------
    Union[Quote,None]

    """
    q = SEARCH_RESULT_DICT.get(keyword)
    # NOTE 兼容旧版本 给缓存加上最后修改时间
    if (
        q is None
        or not q.get("last_time")
        or (
            isinstance(market_type, MarketType)
            and (q.get("classify")) != (market_type.value)
        )
    ):
        return None

    last_time: float = q["last_time"]
    # 缓存过期秒数
    max_ts = 3600 * 24 * 3
    now = time.time()
    # 缓存过期，在线搜索
    if (now - last_time) > max_ts:
        return None
    # NOTE 一定要拷贝 否则改变源对象
    _q = q.copy()
    # NOTE 一定要删除它 否则会构造错误
    del _q["last_time"]
    quote = Quote(**_q)
    return quote


def save_search_result(keyword: str, quotes: List[Quote]):
    """
    存储搜索结果到文件中

    Parameters
    ----------
    keyword : str
        搜索词
    quotes : List[Quote]
        搜索结果
    """
    with open(SEARCH_RESULT_CACHE_PATH, "w", encoding="utf-8") as f:
        # TODO考虑如何存储多个搜索结果
        for quote in quotes:
            now = time.time()
            d = dict(quote._asdict())
            d["last_time"] = now
            SEARCH_RESULT_DICT[keyword] = d
            break
        json.dump(SEARCH_RESULT_DICT.copy(), f)


def rename_dataframe_and_series(
    fields: dict, to_be_removed: List[str] = [], keep_all: bool = True
):
    """
    重命名 DataFrame 和 Series 的列名的装饰器

    Parameters
    ----------
    fields : dict
        新的表头
    to_be_removed : List[str], optional
        要移除的列, by default []
    keep_all : bool, optional
        是否保存全部列(包含未重命名的列), by default True
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            values = func(*args, **kwargs)
            if isinstance(values, pd.DataFrame):
                columns = list(fields.values())
                if keep_all:
                    for column in values.columns:
                        if column not in columns:
                            columns.append(column)
                    values = values.rename(columns=fields)[columns]
                else:
                    values = values.rename(columns=fields)[columns]
                for column in values:
                    if column in to_be_removed:
                        del values[column]
            elif isinstance(values, pd.Series):
                values = values.rename(fields)

            return values

        return wrapper

    return decorator


def process_dataframe_and_series(
    function_fields: Dict[str, Callable] = dict(),
    remove_columns_and_indexes: List[str] = list(),
):
    """
    对 DataFrame 和 Series 进一步操作

    Parameters
    ----------
    function_fields : Dict[str, Callable], optional
        函数字典
    remove_columns_and_indexes : List[str], optional
        需要删除的行或者列, by default list()
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            values = func(*args, **kwargs)
            if isinstance(values, pd.DataFrame):
                for column, function_name in function_fields.items():
                    if column not in values.columns:
                        continue
                    values[column] = values[column].apply(function_name)
                for column in remove_columns_and_indexes:
                    if column in values.columns:
                        del values[column]
            elif isinstance(values, pd.Series):
                for index in remove_columns_and_indexes:
                    values = values.drop(index)
            return values

        return wrapper

    return decorator


T = TypeVar("T")


def to_type(f: Callable[[str], T], value: Any, default: T = None) -> T:
    """
    类型转换

    Parameters
    ----------
    f : Callable[[str], T]
        转换函数
    value : Any
        待转换的值

    default : T, optional
        转换失败时的返回值, 默认为  ``None`` 表示原样返回

    Returns
    -------
    T
        转换结果
    """
    try:
        value = f(value)
        return value
    except:
        if default is None:
            return value
        return default


def add_market(
    category: str, market_number: str, market_name: str, drop_duplicate: bool = True
) -> None:
    """
    添加市场

    Parameters
    ----------
    category : str
        市场类别
    market_number : str
        市场编号
    market_name : str
        市场名称
    drop_duplicate : bool, optional
        是否去重, 默认为 ``True``
    """
    MARKET_NUMBER_DICT[market_number] = market_name
    old = FS_DICT.get(category, "")
    new = f"{old},m:{market_number}"
    if drop_duplicate:
        FS_DICT[category] = ",".join(OrderedDict.fromkeys(new.split(",")))
    else:
        FS_DICT[category] = new


__all__ = []



from pathlib import Path

HERE = Path(__file__).parent

# 债券基本信息表头
EASTMONEY_BOND_BASE_INFO_FIELDS = {
    "SECURITY_CODE": "债券代码",
    "SECURITY_NAME_ABBR": "债券名称",
    "CONVERT_STOCK_CODE": "正股代码",
    "SECURITY_SHORT_NAME": "正股名称",
    "RATING": "债券评级",
    "PUBLIC_START_DATE": "申购日期",
    "ACTUAL_ISSUE_SCALE": "发行规模(亿)",
    "ONLINE_GENERAL_LWR": "网上发行中签率(%)",
    "LISTING_DATE": "上市日期",
    "EXPIRE_DATE": "到期日期",
    "BOND_EXPIRE": "期限(年)",
    "INTEREST_RATE_EXPLAIN": "利率说明",
}



from .getter import (
    get_deal_detail,
    get_futures_base_info,
    get_quote_history,
    get_realtime_quotes,
)

__all__ = [
    "get_futures_base_info",
    "get_quote_history",
    "get_realtime_quotes",
    "get_deal_detail",
]



# from ..stock import get_quote_history
from .getter import (
    get_all_base_info,
    get_base_info,
    get_deal_detail,
    get_history_bill,
    get_quote_history,
    get_realtime_quotes,
    get_today_bill,
)

__all__ = [
    "get_quote_history",
    "get_realtime_quotes",
    "get_all_base_info",
    "get_base_info",
    "get_today_bill",
    "get_history_bill",
    "get_deal_detail",
]



"""
.. include:: ../README.md
"""

__docformat__ = "restructuredtext"
from efinance.api import bond, fund, futures, stock

from efinance import utils


__all__ = ["stock", "fund", "bond", "futures", "utils"]
